(*
 *  COMS W4115
 *  YOLOP
 *
 *  Lisa Li
 *  Jonathan Liu
 *  Sasha McIntosh
 *
 *)

type op = Add | Sub | Mult | Div | Equal | Neq | Less | Leq | Greater | Geq | And | Or

type prim = Int | String | Pixel | Image | Void

type var_decl = {vname: string; vtype: prim}

type expr =
    IntLiteral of int
  | StringLiteral of string
  | Id of string
  | Binop of expr * op * expr
  | Assign of string * expr
  | Call of string * expr list
  | Not of expr
  | ImgPixel of expr * expr * expr * expr 
  | ImgImg of expr * expr
  | Readin of string * expr * expr
  | Writeout of expr * expr * expr
  | Get of expr * expr * expr
  | Set of expr * expr * expr * expr
  | Getc of expr * expr
  | Setc of expr * expr * expr
  | Dimension of expr * string
  | Noexpr

type var_init = 
    Vinit of var_decl * expr

type stmt =
    Block of stmt list
  | Expr of expr
  | Return of expr
  | If of expr * stmt * stmt
  | For of expr * expr * expr * stmt
  | While of expr * stmt
  | Break
  | Continue

type var_def = 
   VarDecl of var_decl
  | Varinit of var_init

type func_decl = {
    fname : string;
    formals : var_decl list;
    locals : var_def list;
    body : stmt list;
    freturn: prim
  }

type program = var_def list * func_decl list

let string_of_vartype = function
   Int -> "int"
 | Void -> "void"
 | String -> "string"
 | Pixel -> "pixel"
 | Image -> "img"

let rec string_of_expr = function
    IntLiteral(l) -> string_of_int l
  | StringLiteral(l) -> l
  | Id(s) -> s
  | Binop(e1, o, e2) ->
      string_of_expr e1 ^ " " ^
      (match o with
	Add -> "+" | Sub -> "-" | Mult -> "*" | Div -> "/"
      | Equal -> "==" | Neq -> "!="
      | Less -> "<" | Leq -> "<=" | Greater -> ">" | Geq -> ">="
      | And -> "&&" | Or -> "||") ^ " " ^
      string_of_expr e2
  | Not(e) -> "!" ^ string_of_expr e
  | ImgPixel(e1, e2, e3, e4) -> "Color(" ^ string_of_expr e1 ^ ", " ^ string_of_expr e2 ^ ", " ^ string_of_expr e3 ^ ", " ^ string_of_expr e4 ^ ");\n" 
  | ImgImg(e1, e2) -> "Image(" ^ string_of_expr e1 ^ ", " ^ string_of_expr e2 ^ "Color(\"white\");\n"
  | Readin(e1, e2, e3) -> "read_in(" ^ e1 ^ ", " ^ string_of_expr e2 ^ ", " ^ string_of_expr e3 ^ ");\n"
  | Writeout(e1, e2, e3) -> "write_out(" ^ string_of_expr e1 ^ ", " ^ string_of_expr e2 ^ ", " ^ string_of_expr e3 ^ ");\n"
  | Get(e1, e2, e3) -> "get(" ^ string_of_expr e1 ^ ", " ^ string_of_expr e2 ^ ", " ^ string_of_expr e3 ^ ");\n"
  | Set(e1, e2, e3, e4) -> "set(" ^ string_of_expr e1 ^ ", " ^ string_of_expr e2 ^ ", " ^ string_of_expr e3 ^ ", " ^ string_of_expr e4 ^ ");\n"
  | Getc(e1, e2) -> "getc(" ^ string_of_expr e1 ^ ", " ^ string_of_expr e2 ^ ");\n"
  | Setc(e1, e2, e3) -> "setc(" ^ string_of_expr e1 ^ ", " ^ string_of_expr e2 ^ ", " ^ string_of_expr e3 ^ ");\n"
  | Dimension(e1, e2) -> "dimension(" ^ string_of_expr e1 ^ ", " ^ e2 ^ ");\n"
  | Assign(v, e) -> v ^ " = " ^ string_of_expr e
  | Call(f, el) ->
      f ^ "(" ^ String.concat ", " (List.map string_of_expr el) ^ ")"
  | Noexpr -> ""

let rec string_of_stmt = function
    Expr(expr) -> string_of_expr expr ^ ";\n"
  | Block(stmts) -> "{\n" ^ String.concat "" (List.map string_of_stmt stmts) ^ "}\n"
  | Return(expr) -> "return " ^ string_of_expr expr ^ ";\n";
  | If(e, s, Block([])) -> "if (" ^ string_of_expr e ^ ")\n {" ^ string_of_stmt s ^ "}"
  | If(e, s1, s2) ->  "if (" ^ string_of_expr e ^ ")\n {" ^
      string_of_stmt s1 ^ "} else\n {" ^ string_of_stmt s2 ^ "}"
  | For(e1, e2, e3, s) ->
      "for (" ^ string_of_expr e1  ^ " ; " ^ string_of_expr e2 ^ " ; " ^
      string_of_expr e3 ^ ")\n{" ^ string_of_stmt s ^ "}"
  | While(e, s) -> "while (" ^ string_of_expr e ^ ")\n {" ^ string_of_stmt s ^ "}"
  | Break -> "break;\n"
  | Continue -> "continue;\n"

let string_of_vartype = function
   Int -> "int"
 | Void -> "void"
 | String -> "string"
 | Pixel -> "Color"
 | Image -> "Image"

(* Have to edit this so that it allows strings as well *)
let string_of_vdecl id = (string_of_vartype id.vtype) ^ " " ^ id.vname

let string_of_vinit = function
   Vinit(v, e) -> string_of_vdecl v ^ " = " ^ string_of_expr e

let string_of_vdef = function
   VarDecl(v) -> string_of_vdecl v ^ ";\n"
 | Varinit(vi) -> string_of_vinit vi ^ ";\n"

let string_of_fdecl fdecl =
  fdecl.fname ^ " (" ^ String.concat ", " (List.map string_of_vdecl fdecl.formals) ^ ")\n{\n" ^ 
  (String.concat "" (List.map string_of_vdef fdecl.locals)) ^
  String.concat "" (List.map string_of_stmt fdecl.body) ^
  "}\n"

(* Might need to edit this as well? Dunno. *)
let string_of_program (vars, funcs) =
  String.concat "" (List.map string_of_vdef vars) ^ "\n" ^
  String.concat "\n" (List.map string_of_fdecl funcs)
